import socket
import struct
import ssl
import usocket as socket
import random
import binascii
import machine
from websocket import websocket


class WebSocketClient(websocket):
    def __init__(self, sock):
        super().__init__(sock)
        self.__sock = sock

    def setblocking(self, value):
        self.__sock.setblocking(value)

    @staticmethod
    def create_ssl_socket(hostname):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        addr = socket.getaddrinfo(hostname, 443)[0][-1]
        s.connect(addr)
        return ssl.wrap_socket(s)

    @staticmethod
    def create_websocket(hostname, path, protocol=None):
        sock = WebSocketClient.create_ssl_socket(hostname)

        query = f"GET {path} HTTP/1.1\r\n"
        query += f"Host: {hostname}\r\n"
        query += f"Connection: Upgrade\r\n"
        query += f"Upgrade: websocket\r\n"
        query += f"Sec-WebSocket-Key: {binascii.b2a_base64(bytes(random.getrandbits(8)
                                                                 for _ in range(16)))[:-1].decode('utf-8')}\r\n"
        query += f"Sec-WebSocket-Version: 13\r\n"
        if protocol is not None:
            query += f"Sec-WebSocket-Protocol: {protocol}\r\n"
        query += f"\r\n"

        sock.write(query.encode('utf-8'))

        header = sock.readline()[:-2]
        if not header:
            sock.close()
            print("ERROR: WebSocket not created (no response)")
            return None

        if not header.startswith(b'HTTP/1.1 101 '):
            sock.close()
            print(f"ERROR: WebSocket not created (server returns '{header.decode('utf-8')}')")
            return None

        print(f"INFO: WebSocket created! (server response: {header.decode('utf-8')})")

        while header:
            if header == b'\r\n':
                break
            else:
                print(header)
            header = sock.readline()

        return WebSocketClient(sock)


class WSSMQTTException(Exception):
    pass


class WSSMQTTClient:
    def __init__(
            self,
            host,
            path='/',
            client_id=f"client-{machine.unique_id().hex()}",
            user=None,
            password=None,
            keepalive=0,
    ):
        self.client_id = client_id
        self.sock = None
        self.host = host
        self.path = path
        self.pid = 0
        self.cb = None
        self.user = user
        self.pswd = password
        self.keepalive = keepalive
        self.lw_topic = None
        self.lw_msg = None
        self.lw_qos = 0
        self.lw_retain = False

    def _send_str(self, s):
        self.sock.write(struct.pack("!H", len(s)))
        self.sock.write(s)

    def _recv_len(self):
        n = 0
        sh = 0
        while 1:
            b = self.sock.read(1)[0]
            n |= (b & 0x7F) << sh
            if not b & 0x80:
                return n
            sh += 7

    def set_callback(self, f):
        self.cb = f

    def set_last_will(self, topic, msg, retain=False, qos=0):
        assert 0 <= qos <= 2
        assert topic
        self.lw_topic = topic
        self.lw_msg = msg
        self.lw_qos = qos
        self.lw_retain = retain

    def connect(self, clean_session=True):
        if self.sock is not None:
            raise WSSMQTTException('Already connected')

        self.sock = WebSocketClient.create_websocket(self.host, self.path, protocol='mqtt')
        if self.sock is None:
            raise WSSMQTTException('Failed to create websocket')

        try:
            premsg = bytearray(b"\x10\0\0\0\0\0")
            msg = bytearray(b"\x04MQTT\x04\x02\0\0")

            sz = 10 + 2 + len(self.client_id)
            msg[6] = clean_session << 1
            if self.user:
                sz += 2 + len(self.user) + 2 + len(self.pswd)
                msg[6] |= 0xC0
            if self.keepalive:
                assert self.keepalive < 65536
                msg[7] |= self.keepalive >> 8
                msg[8] |= self.keepalive & 0x00FF
            if self.lw_topic:
                sz += 2 + len(self.lw_topic) + 2 + len(self.lw_msg)
                msg[6] |= 0x4 | (self.lw_qos & 0x1) << 3 | (self.lw_qos & 0x2) << 3
                msg[6] |= self.lw_retain << 5

            i = 1
            while sz > 0x7F:
                premsg[i] = (sz & 0x7F) | 0x80
                sz >>= 7
                i += 1
            premsg[i] = sz

            self.sock.write(premsg, i + 2)
            self.sock.write(msg)
            # print(hex(len(msg)), hexlify(msg, ":"))
            self._send_str(self.client_id)
            if self.lw_topic:
                self._send_str(self.lw_topic)
                self._send_str(self.lw_msg)
            if self.user:
                self._send_str(self.user)
                self._send_str(self.pswd)
            resp = self.sock.read(4)
            assert resp[0] == 0x20 and resp[1] == 0x02
            if resp[3] != 0:
                raise WSSMQTTException(resp[3])
            return resp[2] & 1
        except BaseException as e:
            self.sock.close()
            self.sock = None
            raise e

    def disconnect(self):
        if self.sock is None:
            return
        self.sock.write(b"\xe0\0")
        self.sock.close()
        self.sock = None

    def is_connected(self):
        return self.sock is not None

    def ping(self):
        self.sock.write(b"\xc0\0")

    def publish(self, topic, msg, retain=False, qos=0):
        pkt = bytearray(b"\x30\0\0\0")
        pkt[0] |= qos << 1 | retain
        sz = 2 + len(topic) + len(msg)
        if qos > 0:
            sz += 2
        assert sz < 2097152
        i = 1
        while sz > 0x7F:
            pkt[i] = (sz & 0x7F) | 0x80
            sz >>= 7
            i += 1
        pkt[i] = sz
        # print(hex(len(pkt)), hexlify(pkt, ":"))
        self.sock.write(pkt, i + 1)
        self._send_str(topic)
        if qos > 0:
            self.pid += 1
            pid = self.pid
            struct.pack_into("!H", pkt, 0, pid)
            self.sock.write(pkt, 2)
        self.sock.write(msg)
        if qos == 1:
            while 1:
                op = self.wait_msg()
                if op == 0x40:
                    sz = self.sock.read(1)
                    assert sz == b"\x02"
                    rcv_pid = self.sock.read(2)
                    rcv_pid = rcv_pid[0] << 8 | rcv_pid[1]
                    if pid == rcv_pid:
                        return
        elif qos == 2:
            assert 0

    def subscribe(self, topic, qos=0):
        assert self.cb is not None, "Subscribe callback is not set"
        pkt = bytearray(b"\x82\0\0\0")
        self.pid += 1
        struct.pack_into("!BH", pkt, 1, 2 + 2 + len(topic) + 1, self.pid)
        # print(hex(len(pkt)), hexlify(pkt, ":"))
        self.sock.write(pkt)
        self._send_str(topic)
        self.sock.write(qos.to_bytes(1, "little"))
        while 1:
            op = self.wait_msg()
            if op == 0x90:
                resp = self.sock.read(4)
                # print(resp)
                assert resp[1] == pkt[2] and resp[2] == pkt[3]
                if resp[3] == 0x80:
                    raise WSSMQTTException(resp[3])
                return

    # Wait for a single incoming MQTT message and process it.
    # Subscribed messages are delivered to a callback previously
    # set by .set_callback() method. Other (internal) MQTT
    # messages processed internally.
    def wait_msg(self):
        res = self.sock.read(1)
        if res is None:
            return None
        if res == b"":
            raise OSError(-1)
        self.sock.setblocking(True)
        if res == b"\xd0":  # PINGRESP
            sz = self.sock.read(1)[0]
            assert sz == 0
            return None
        op = res[0]
        if op & 0xF0 != 0x30:
            return op
        sz = self._recv_len()
        topic_len = self.sock.read(2)
        topic_len = (topic_len[0] << 8) | topic_len[1]
        topic = self.sock.read(topic_len)
        sz -= topic_len + 2
        if op & 6:
            pid = self.sock.read(2)
            pid = pid[0] << 8 | pid[1]
            sz -= 2
        msg = self.sock.read(sz)
        self.cb(topic, msg)
        if op & 6 == 2:
            pkt = bytearray(b"\x40\x02\0\0")
            struct.pack_into("!H", pkt, 2, pid)
            self.sock.write(pkt)
        elif op & 6 == 4:
            assert 0
        return op

    # Checks whether a pending message from server is available.
    # If not, returns immediately with None. Otherwise, does
    # the same processing as wait_msg.
    def check_msg(self):
        self.sock.setblocking(False)
        return self.wait_msg()

